/*
 * t32 -- convert t32 files into assembler
 */
/* Author:	Rob Pike  (Bell Labs)  Summer 1980
 * Revise:	David Shields  (NYU-CIMS)  13-Nov-81
 *		add mvw, lha and sha opcodes
 * Revise:	David Shields  (NYU-CIMS)  4-Jan-82
 *		support separate compilation.
 *		drop -g switch ; do not make "gN" symbols global.
 * Revise:      Steve Knodle (Clarkson)   29-Mar-83
 *		add ".v" switch to allow assembly language
 *		fragments to be passed through verbatim.
 *		Affected: getline
*/
/*
 * Meaning of the codes:
 *
 *	put	output the line
 *	if2i	output the line if the 2nd arg is an immediate mode quantity
 *	if2ni	output the line if the 2nd arg is NOT immediate mode
 *	mask	allocate a mask symbol (M123 etc.)
 *	nlab	code sequence uses number of labels given by char string
 *	lbl	make sure that stated argument has 'l' turned into 'L'
 *	mlbl	stated argument may need to be lbl'd
 *	ifdif	output line if first 2 arguments are not the same
 *	ifsam	output line if first 2 arguments are the same
 *	dolbar	turn '$' into '_' in stated argument
 *	text	move into named text space
 *	data	move into named data space
 *	mark	prepare a ".org" statement from stated arguments which will
 *		be output at end of current data definition
 *	ifmain	change arg 1 to specified string if this is main proc
 *	split	create new output file named <specified arg>.s
 *
 *	In the output string, A2 gets turned into the second argument,
 *	M gets turned into the name of the current mask symbol,
 *	and L1 gets turned into the current label symbol plus 1.
 */
int	put(), if2i(), if2ni(), mask(), nlab(), 
		lbl(), ifdif(), ifsam(), dolbar();
int	text(), data(), mark(), mlbl(), ifmain(), split();
char zero[]="0";
char one[]="1";
#define	TEXT	0
#define	DATA	1
int	textdata=TEXT;
char	tdspace='0';	/* which text or data space we are in */
char	markbuf[128];	/* buffer to store .org to mark end of data area */
struct code{
	int	(*op)();
	char	*optext;
};
#define CODE(op)	struct code op[]
CODE(ban)={
	text,	zero,
	if2i,	"	bicl2	A2^0xFFFFFFFF,A1",
	if2ni,	"	mcoml	A2,r1",
	if2ni,	"	bicl2	r1,A1",
	0,	0,
};
CODE(bfb)={
	text,	zero,
	nlab,	"2",
	put,	"	movl	$31,r1",
	put,	"L0:	bbs	r1,A2,L1",
	put,	"	sobgeq	r1,L0",
	put,	"L1:	addl3	$1,r1,A1",
	0,	0,
};
CODE(bnb)={
	text,	zero,
	nlab,	"2",
	put,	"	clrl	r1",
	put,	"	clrl	-(sp)",
	put,	"L0:	bbc	r1,A2,L1",
	put,	"	incl	(sp)",
	put,	"L1:	acbl	$31,$1,r1,L0",
	put,	"	movl	(sp)+,A1",
	0,	0,
};
CODE(bno)={
	text,	zero,
	put,	"	mcoml	A2,A1",
	0,	0,
};
CODE(bor)={
	text,	zero,
	put,	"	bisl2	A2,A1",
	0,	0,
};
CODE(bxo)={
	text,	zero,
	put,	"	xorl2	A2,A1",
	0,	0,
};
CODE(cal)={
	text,	zero,
	dolbar,	"1",
	put,	"	callg	A3,A1",
	0,	0,
};
CODE(ceq)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A1,A2",
	put,	"L0:	beql	L0+2+3",
	0,	0,
};
CODE(cge)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A1,A2",
	put,	"L0:	bgeq	L0+2+3",
	0,	0,
};
CODE(cgt)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A2,A1",
	put,	"L0:	blss	L0+2+3",
	0,	0,
};
CODE(cle)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A2,A1",
	put,	"L0:	bgeq	L0+2+3",
	0,	0,
};
CODE(clt)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A1,A2",
	put,	"L0:	blss	L0+2+3",
	0,	0,
};
CODE(cne)={
	text,	zero,
	nlab,	"1",
	put,	"	cmpl	A1,A2",
	put,	"L0:	bneq	L0+2+3",
	0,	0,
};
CODE(dbr)={
	text,	one,
	mark,	"12",
	nlab,	"1",
	put,	"A1:",
	0,	0,
};
CODE(dbw)={
	data,	zero,
	mark,	"12",
	nlab,	"1",
	put,	"A1:",
	0,	0,
};
CODE(dec)={
	text,	zero,
	put,	"	.set	M,0xA2<2",
	0,	0,
};
CODE(dep)={
	text,	zero,
	0,	0,
};
CODE(dna)={
	0,	0,
};
CODE(dnd)={
	data,	one,
	mark,	"23",
	dolbar,	"1",
	put,	"	.globl	A1",
	put,	"A1:",
	0,	0,
};
CODE(dsc)={
	text,	zero,
	0,	0,
};
CODE(dsp)={
	split,	"1",
	nlab,	"1",
	put,	"#start	A1",
	put,	"	.text	0",
	text,	zero,
	mask,	0,
	dolbar,	"1",
	ifmain,	"__ltlmai",
	put,	".globl	A1",
	put,	"A1:",
	put,	"	.word	M",
	put,	"	movab	L0,r0",
	put,	"	jsb	mcount",
	put,	"	.data",
	put,	"	.align 	2",
	put,	"L0:	.long	0",
	put,	"	.text",
	0,	0,
};
CODE(dwa)={
	mlbl,	"2",
	put,	"	.org	A1",
	put,	"	.long	A2",
	0,	0,
};
CODE(dwc)={
	put,	"	.org	A1",
	put,	"	.byte	'A2",
	0,	0,
};
CODE(dwh)={
	put,	"	.org	A1",
	put,	"	.long	0xA2",
	0,	0,
};
CODE(dwi)={
	put,	"	.org	A1",
	put,	"	.long	A2",
	0,	0,
};
CODE(dwo)={
	put,	"	.org	A1",
	put,	"	.long	0A2",
	0,	0,
};
CODE(dwr)={
	put,	"	.org	A1",
	put,	"	.long	0fA2",
	0,	0,
};
CODE(dws)={
	put,	"	.org	A1",
	put,	"	.ascii	A2",
	0,	0,
};
CODE(dwz)={
	put,	"	.org	A1",
	put,	"	.space	4*A2",
	0,	0,
};
CODE(ent)={
	text,	zero,
	0,	0,
};
CODE(iab)={
	text,	zero,
	nlab,	"1",
	ifdif,	"	movl	A2,A1",
	ifsam,	"	tstl	A1",
	put,	"	bgeq	L0",
	put,	"	mnegl	A2,A1",
	put,	"L0:",
	0,	0,
};
CODE(iad)={
	text,	zero,
	put,	"	addl2	A2,A1",
	0,	0,
};
CODE(iao)={
	text,	zero,
	ifdif,	"	movl	A2,A1",
	put,	"	incl	A1",
	0,	0,
};
CODE(ico)={
	text,	zero,
	put,	"	mnegl	A2,A1",
	0,	0,
};
CODE(idi)={
	text,	zero,
	put,	"	divl2	A2,A1",
	0,	0,
};
CODE(idt)={
	text,	zero,
	if2i,	"	ashl	A2^0xFFFFFFFF+1,A1,A1",
	if2ni,	"	mnegl	A2,r1",
	if2ni,	"	ashl	r1,A1,A1",
	0,	0,
};
CODE(ieq)={
	text,	zero,
	nlab,	"1",
	put,	"	xorl2	A2,A1",
	put,	"	beql	L0",
	put,	"	mnegl	$1,A1",
	put,	"L0:	incl	A1",
	0,	0,
};
CODE(ige)={
	text,	zero,
	put,	"	subl2	A2,A1",
	put,	"	ashl	$-31,A1,A1",
	put,	"	incl	A1",
	0,	0,
};
CODE(ilt)={
	text,	zero,
	put,	"	subl2	A2,A1",
	put,	"	mcoml	A1,A1",
	put,	"	ashl	$-31,A1,A1",
	put,	"	incl	A1",
	0,	0,
};
CODE(imo)={
	text,	zero,
	put,	"	divl3	A2,A1,r1",
	put,	"	mull2	A2,r1",
	put,	"	subl2	r1,A1",
	0,	0,
};
CODE(imt)={
	text,	zero,
	put,	"	ashl	A2,A1,A1",
	0,	0,
};
CODE(imu)={
	text,	zero,
	put,	"	mull2	A2,A1",
	0,	0,
};
CODE(ine)={
	text,	zero,
	nlab,	"1",
	put,	"	xorl2	A2,A1",
	put,	"	beql	L0",
	put,	"	movl	$1,A1",
	put,	"L0:",
	0,	0,
};
CODE(isi)={
	text,	zero,
	nlab,	"2",
	put,	"	tstl	A1",
	put,	"	bgeq	L0",
	put,	"	mnegl	A1,A1",
	put,	"L0:	tstl	A2",
	put,	"	bgeq	L1",
	put,	"	mnegl	A1,A1",
	put,	"L1:",
	0,	0,
};
CODE(iso)={
	text,	zero,
	ifdif,	"	movl	A2,A1",
	put,	"	decl	A1",
	0,	0,
};
CODE(isu)={
	text,	zero,
	put,	"	subl2	A2,A1",
	0,	0,
};
CODE(jeq)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	bneq	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(jgt)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	bleq	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(jge)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	blss	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(jle)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	bgtr	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(jlt)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	bgeq	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(jma)={
	text,	zero,
	lbl,	"2",
	put,	"	brw	A2",
	0,	0,
};
CODE(jne)={
	text,	zero,
	nlab,	"1",
	lbl,	"2",
	put,	"	tstl	A1",
	put,	"	beql	L0",
	put,	"	brw	A2",
	put,	"L0:",
	0,	0,
};
CODE(lab)={
	text,	zero,
	lbl,	"1",
	put,	"A1:",
	0,	0,
};
CODE(lda)={
	text,	zero,
	put,	"	moval	A2,A1",
	0,	0,
};
CODE(ldw)={
	text,	zero,
	put,	"	movl	A2,A1",
	0,	0,
};
CODE(lha)={
	text,	zero,
	put,	"	movl	A3,A1",
	put,	"	ashl	$-2,A1,A2",
	0,	0,
};
CODE(lpr)={
	text,	zero,
	if2i,	"	extzv	A3&31,A4,A2,A1",
	if2ni,	"	extzv	A3,A4,A2,A1",
	0,	0,
};
CODE(mvw)={
	text,	zero,
	put,	"	pushr	$077",
	put,	"	movc3	A3*4,A2,(A1)",
	put,	"	popr	$077",
	0,	0,
};
CODE(mvx)={
	text,	zero,
	put,	"	pushr	$077",
	put,	"	movc3	A3*4,(A1),A2",
	put,	"	popr	$077",
	0,	0,
};
CODE(sha)={
	text,	zero,
	put,	"	moval	A1,A2",
	0,	0,
};
CODE(spr)={
	text,	zero,
	if2i,	"	insv	A1,A3&31,A4,A2",
	if2ni,	"	insv	A1,A3,A4,A2",
	0,	0,
};
CODE(stw)={
	text,	zero,
	put,	"	movl	A1,A2",
	0,	0,
};
CODE(xjm)={
	text,	zero,
	put,	"	jmp	*A2(A1)",
	0,	0,
};
CODE(zeb)={
	text,	zero,
	put,	"	pushr	$077",
	put,	"	movc5	$0,(A1),$0,A2*4,(A1)",
	put,	"	popr	$077",
	0,	0,
};
CODE(zew)={
	text,	zero,
	put,	"	clrl	A2",
	0,	0,
};
CODE(rdi)={
	text,	zero,
	put,	"	divf2	A2,A1",
	0,	0,
};
CODE(ret)={
	text,	zero,
	put,	"	ret",
	0,	0,
};
CODE(rmu)={
	text,	zero,
	put,	"	mulf2	A2,A1",
	0,	0,
};
CODE(rad)={
	text,	zero,
	put,	"	addf2	A2,A1",
	0,	0,
};
CODE(rsu)={
	text,	zero,
	put,	"	subf2	A2,A1",
	0,	0,
};
CODE(ifr)={
	text,	zero,
	put,	"	cvtfl	A2,A1",
	0,	0,
};
CODE(rfi)={
	text,	zero,
	put,	"	cvtlf	A2,A1",
	0,	0,
};
CODE(rlt)={
	text,	zero,
	nlab,	"1",
	put,	"	clrl	r1",
	put,	"	cmpf	A1,A2",
	put,	"	bgeq	L0",
	put,	"	incl	r1",
	put,	"L0:	movl	r1,A1",
	0,	0,
};
CODE(rgt)={
	text,	zero,
	nlab,	"1",
	put,	"	clrl	r1",
	put,	"	cmpf	A1,A2",
	put,	"	bleq	L0",
	put,	"	incl	r1",
	put,	"L0:	movl	r1,A1",
	0,	0,
};
CODE(rge)={
	text,	zero,
	nlab,	"1",
	put,	"	clrl	r1",
	put,	"	cmpf	A1,A2",
	put,	"	blss	L0",
	put,	"	incl	r1",
	put,	"L0:	movl	r1,A1",
	0,	0,
};
CODE(rle)={
	text,	zero,
	nlab,	"1",
	put,	"	clrl	r1",
	put,	"	cmpf	A1,A2",
	put,	"	bgtr	L0",
	put,	"	incl	r1",
	put,	"L0:	movl	r1,A1",
	0,	0,
};
CODE(rab)={
	text,	zero,
	nlab,	"1",
	put,	"	movf	A2,A1",
	put,	"	bgeq	L0",
	put,	"	mnegf	A1,A1",
	put,	"L0:",
	0,	0,
};
CODE(rco)={
	text,	zero,
	ifdif,	"	movf	A2,A1",
	put,	"	mnegf	A1,A1",
	0,	0,
};
CODE(null)={
	0,	0,
};
/*
 * The entries in the macro table are in decreasing order of calling frequency
 */
struct macro{
	char		*macname;
	struct code	*maccode;
}macro[]={
	"",	null,
	"ldw",	ldw,	"dwh",	dwh,	"dwa",	dwa,	"imt",	imt,	"cal",	cal,
	"lha",	lha,	"lpr",	lpr,	
	"stw",	stw,	"lab",	lab,	"dwi",	dwi,	"jma",	jma,
	"lda",	lda,	"sha",	sha,
	"spr",	spr,	"jeq",	jeq,	"iao",	iao,	"zew",	zew,
	"imu",	imu,	"dbw",	dbw,	"ceq",	ceq,	"jne",	jne,	"dwz",	dwz,
	"iso",	iso,	"isu",	isu,	"iad",	iad,	"cne",	cne,	"dna",	dna,
	"ent",	ent,	"end",	null,	"dsp",	dsp,	"dsc",	dsc,	"dep",	dep,
	"dec",	dec,	"ret",	ret,	"cgt",	cgt,	"mvw",	mvw,	
	"mvx",	mvx,	"zeb",	zeb,
	"dbr",	dbr,	"ban",	ban,	"bor",	bor,	"clt",	clt,	"idt",	idt,
	"ine",	ine,	"jle",	jle,	"cle",	cle,	"ieq",	ieq,	"cge",	cge,
	"ico",	ico,	"jgt",	jgt,	"xjm",	xjm,	"jge",	jge,	"ilt",	ilt,
	"iab",	iab,	"bxo",	bxo,	"bno",	bno,	"bfb",	bfb,	"dnd",	dnd,
	"bnb",	bnb,	"jlt",	jlt,	"imo",	imo,	"idi",	idi,	"dwc",	dwc,
	"dwo",	dwo,	"dwr",	dwr,	"dws",	dws,	"ifr",	ifr,	"ige",	ige,
	"isi",	isi,	"rab",	rab,	"rad",	rad,	"rco",	rco,	"rdi",	rdi,
	"req",	ieq,	"rfi",	rfi,	"rge",	rge,	"rgt",	rgt,	"rle",	rle,
	"rlt",	rlt,	"rmu",	rmu,	"rne",	ine,	"rsu",	rsu,
	0,	0,
};

#include <stdio.h>
char	line[128];
char	argline[128];
int	debug=0;
int	splitf=0;
int	defg=1;
long	lineno=0L;
long	labnum=0L;
long	masknum=0L;
long	nextlab=0L;
int	ac;
char	*av[20];
struct code *
lookup(s)
	register char *s;
{
	register struct macro *mp;
	for(mp=macro; mp->macname; mp++)
		if(strcmp(s, mp->macname)==0)
			return(mp->maccode);
	fprintf(stderr, "t32: line %d: %s -- unknown operand %s\n",
			lineno, line, s);
abort();
	exit(1);
}
main(argc, argv)
	char *argv[];
{
	register i;
	while(argc>1 && argv[1][0]=='-'){
		switch(argv[1][1]){
		default:
			fprintf(stderr, "t32: unknown flag %s\n", argv[1]);
			fprintf(stderr, "usage: t32 [-d] [-g] [-s] [file]\n");
			exit(1);
		case 'd':
			debug++;
			break;
		case 's':
			splitf++;
			break;
		}
		argv++; --argc;
	}
	while(getline()){
		argify();
		expand(lookup(av[0]));
		labnum=nextlab;
	}
}
getline(){

/* this procedure contains CCT local modifications to allow
   modifications to ltlasm to pass through. 
   input lines beginning ".v" cause succeeding lines
   to be copied verbatim until another line beginning ".v"     
   is received.*/

#define block while((c=getchar())!='\n' && c!=';' && c!=EOF) *p++=c;
#define next_line  p = line; block *p='\0';

#define verbatim_test *p++ == '.' && *p == 'v' 

	register char c, *p;
	next_line;
	if(c==EOF)
		return(0);
	*p='\0';
	if(debug){
		putchar('#');
		puts(line);
	}
	/* local mods begin here                         */

        p = line;
        if(verbatim_test) {
	     next_line;
	     do {
                puts(line);		  
		next_line;
	        p = line;
	     }
	     while (!(verbatim_test));
	     next_line;
        }
	return(1);
}
char *
skipbl(p)
	register char *p;
{
	while(*p==' ' || *p=='\t' || *p==',')
		p++;
	return(p);
}
argify(){
	register char *p, *q;
	register i;
	ac=0;
	av[0]="";
	p=line;
	q=argline;
	while(*(p=skipbl(p))){
		if(*p=='#')
			break;
		av[ac++]=q;
		while((*q++= *p) && *p!=' ' && *p!='\t' && *p!=',')
			p++;
		q[-1]='\0';
	}
	/* Remove spurious +'s */
	for(i=1; i<ac; i++){	/* start at i=1 as av[0] will be macro */
		p=av[i];
		if((p[0]=='$' || p[0]=='*') && p[1]=='+')
			strcpy(p+1, p+2);
		else if(p[0]=='+')
			strcpy(p, p+1);
	}
}
put(s)
	register char *s;
{
	while(*s){
		switch(*s){
		case 'A':
			fputs(av[*++s-'0'], stdout);
			break;
		case 'M':
			printf("M%ld", masknum);
			break;
		case 'L':
			printf("LL%ld", labnum+(*++s-'0'));
			break;
		default:
			putchar(*s);
		}
		s++;
	}
	putchar('\n');
}
if2i(s)
	char *s;
{
	if(av[2][0]=='$')	/* immediate constant */
		put(s);
}
if2ni(s)
	char *s;
{
	if(av[2][0]!='$')	/* something else */
		put(s);
}
mask(s)
	char *s;
{
	masknum++;
}
nlab(s)
	char *s;
{
	nextlab+=s[0]-'0';
}
lbl(s)
	char *s;
{
	if(av[s[0]-'0'][0]!='l'){
		fprintf(stderr, "t32: line %d: %s -- funny label %s\n",
			lineno, line, av[s[0]-'0']);
		exit(1);
	}
	av[s[0]-'0'][0]='L';
}
mlbl(s)
	char *s;
{
	register char *a;
	a=av[s[0]-'0'];
	if(a[0]=='l' && '0'<=a[1] && a[1]<='9')
		a[0]='L';
}
ifdif(s)
	char *s;
{
	if(strcmp(av[1], av[2]))	/* different */
		put(s);
}
ifsam(s)
	char *s;
{
	if(strcmp(av[1], av[2])==0)	/* same */
		put(s);
}
dolbar(s)
	char *s;
{
	s=av[s[0]-'0'];
	do
		if(*s=='$')
			*s='_';
	while(*s++);
}
text(s)
	register char *s;
{
	origin();
	if(textdata!=TEXT || tdspace!=s[0]){
		printf("	.text	%c\n",	s[0]);
		textdata=TEXT;
		tdspace=s[0];
	}
}
data(s)
	register char *s;
{
	origin();
	if(textdata!=DATA || tdspace!=s[0]){
		printf("	.data	%c\n",	s[0]);
		textdata=DATA;
		tdspace=s[0];
	}
}
mark(s)
	char *s;
{
	sprintf(markbuf, "\t.org\t%s+%s\n", av[s[0]-'0'], av[s[1]-'0']);
}
origin(){
	printf(markbuf);
	markbuf[0]='\0';
}
ifmain(s)
	char *s;
{
	if(av[3][0]=='2')
		av[1]=s;
}
split(s)
	char *s;
{
	if(splitf){
		char buf[256];
		if(freopen(strcat(strcpy(buf, av[s[0]-'0']), ".s"), "w", stdout)
			== NULL){
			fprintf(stderr, "t32: can't create %s.s\n",
				av[s[0]-'0']);
			exit(1);
		}
	}
}
expand(cp)
	register struct code *cp;
{
	while(cp->op){
		(*cp->op)(cp->optext);
		cp++;
	}
}
