/*	LITTLE env: 'c' part
 * LAST CHANGE:	13-dec-81 shields
 * This file contains that part of the LITTLE env (environment interface)
 * that can be written in c. The author is
 *	Rob Pike (RP), Bell Labs, Murrary Hill, NJ
 * with changes/extensions added by
 *	David Shields (DS),  CIMS-NYU
 *
 *	Cast:
 *	RP Rob Pike, Bell Labs, original author, summer 1980
 *	DS Dave Shields, NYU-CIMS, maintainer, fall 1981+
 *
 * CHANGES/BUGS (reporter noted, so he can be consulted for details)
 *	(entries starting with BUG describe outstanding bugs.)
 *
 *	DS 21-apr-82
 *		Bypass tty code to open /dev/tty and usage in remarkl
 *	DS 14-jan-82
 *		Revise putcsio and putwsio to change little's fortran
 *		carriage control for print files to unix convention,
 *		emitting formfeed to indicate start of page.
 *	DS 12-jan-81
 *		Use correct defaults of 'stdin' and 'stdout' for
 *		default opens of 1 and 2 with null name, respectively.
 *		Avoid close of std input file; let c library do it.
 *	DS 26-dec-81
 *		Fix sysfin handling of completion codes to reflect
 *		little's error encoding.
 *		In particular, deal with 8-bit limitation of unix.
 *		See code at start of ltllib.opl for summary of error
 *		codes used by little's library.
 *	DS 21-dec-81
 * 		Revise to use 'stdin', 'stdout' and 'stderr' as the 'names'
 *		of i=, l= and term= files.
 *	DS 13-dec-81
 *		BUG: header declarations shared by setl/mk/stlenv.c.
 *		there should be a single declaration text included
 *		in this file and stlenv.c
 *	DS 10-dec-81
 *		BUG: doesn't check that file number in range
 *	DS 10-dec-81
 *		BUG: doesn't check that file connected (open) before steaming on
 *	DS 10-dec-81
 *		BUG: promsio doesn't check for big (>64 chars) prompt
 *	DS 10-dec-81
 *		NOTE: VMS generous in using 'sys$' names to permit multiple files
 *		associated with same name; UNIX is less permissive. As
 *		a consequence, programs (notably Ada compiler) using
 *		'sys$' names prematurely closing files leaving underlying
 *		interface confused.
 *
*/

int tty;
/* declarations below duplicated in part in setl/mk stlenv.c - be warned! */
#include <stdio.h>
#include <sys/types.h>
typedef	int	SDS;
typedef	int	PACK;
typedef	int	UNPACK;
#define	GET	1
#define	PRINT	2
#define	PUT	3
#define	READ	4
#define	STRING	5
#define	WRITE	6
#define	NFILE	20
#define PROMPTSIZE	64
typedef struct{
	FILE	*fptr;
	short	linesize;
	short	errlevel;
	short	access;
	char	name[64];
	short	prompting;
	char	prompt[PROMPTSIZE]; /* if prompting */
}IO;
IO	IOtab[NFILE+1];		/* IOtab[0] is unused */

ltlsio(n)
	int *n;
{
	register i;
	if(*n!=0){
		for(i=1; i<NFILE+1; i++){
			if(IOtab[i].fptr)
				fclose(IOtab[i].fptr);
			IOtab[i].fptr=NULL;
		}
	}
}

IOstdfile(p)
	register IO *p;
{
	register FILE *id;
	if(IOequal(p->name, "stdin"))
		id = stdin;
	else if(IOequal(p->name, "stdout"))
		id = stdout;
	else if(IOequal(p->name, "stderr"))
		id = stderr;
	else
		return(0);
	p->name[0] = '\0';
	p->fptr = id;
	return(1);
}

IOequal(s, t)
	register char *s, *t;
{
	while(*s++ == *t)
		if(*t++ == '\0')
			return(1);
	return(0);
}
opensio(fn, rc, ac, titl, lns, lnsret)
	int *fn, *rc, *ac, *lns, *lnsret;
	SDS *titl;
{
	register IO *p;
	char *acode;
	if(*fn<1 || *fn>NFILE){
  Error:
		*rc=1;
		return;
	}
	p = &IOtab[*fn];
	if(p->fptr)
		goto Error;
	IOsdstoc(titl, p->name);
	if (p->name[0]=='\0')
			if (*fn==1)
				sprintf(p->name, "stdin");
			else if (*fn==2)
				sprintf(p->name, "stdout");
	if(!IOstdfile(p)){
		if(p->name[0]=='\0')
			sprintf(p->name, "file%d", *fn);
		else if(p->name[0]=='0' && p->name[1]=='\0')
			strcpy(p->name, "/dev/null");
		acode = "r";
		if(*ac==PRINT || *ac==PUT || *ac==WRITE)
			acode = "w";
		p->fptr = fopen(p->name, acode);
		if(p->fptr == NULL)
			goto Error;
	}
	*lnsret = p->linesize = *lns==0? 133 : *lns;
	*rc = 0;
	p->access = *ac;
	p->prompting = 0; /* assume no prompting */
}


clossio(fn, rc)
	int *fn, *rc;
{
	*rc = 0;
	if (*fn>1)
	*rc = fclose(IOtab[*fn].fptr)==EOF;
	IOtab[*fn].fptr = NULL;
}

dropsio(fn, rc)
	int *fn, *rc;
{
	*rc = unlink(IOtab[*fn].name);
}

rewisio(fn, rc, ac)
	int *fn, *rc, *ac;
{
	if(*ac==0)
		rewind(IOtab[*fn].fptr);
	else
		(void) freopen(IOtab[*fn].name, "r", IOtab[*fn].fptr);
	*rc = 0;
}

char	IOpbuf[256];
UNPACK	IOupbuf[256];

putcsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	UNPACK *ara;
{
	int n;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	if(IOtab[*fn].fptr==0){
		fprintf(stderr, "fptr=0 in putcsio (fn=%d), write of %s\n", 
			*fn, IOpbuf);
		*rc=0;
	}
	if (IOtab[*fn].access==PRINT) {
		if (ara[*lo-1]==' ') {
			n -= 1;
			IOpack(ara+*lo, IOpbuf, n); }
		else if (ara[*lo-1]=='1') {
			IOpbuf[0] = '\f';
			IOpack(ara+*lo, IOpbuf+1, n-1);}
		else	IOpack(ara+*lo-1, IOpbuf, n);
	}
	else 	IOpack(ara+*lo-1, IOpbuf, n);
	IOstripbl(IOpbuf, n);
	fputs(IOpbuf, IOtab[*fn].fptr);
	fflush(IOtab[*fn].fptr);  
}

putwsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	PACK *ara;
{
	int n, i;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	i = 0;
	IOscramble(ara+*lo-1, IOpbuf, n);
 	/* check for print access, with ff processing */
	if (IOtab[*fn].access==PRINT && n>0) {
		if (IOpbuf[0]=='1')  IOpbuf[0] = '\f';
		else if (IOpbuf[0]==' ')  {
			i = 1; n -= 1;
		}
	}
	IOstripbl(IOpbuf+i, n);
	*rc = fputs(IOpbuf+i, IOtab[*fn].fptr)==EOF;
}

wtrwsio(fn, rc, ara, lo, nw)
	int *fn, *rc, *ara, *lo, *nw;
{
	*rc = 0==fwrite(ara+*lo-1, sizeof(*ara), *nw, IOtab[*fn].fptr);
}

rdrwsio(fn, rc, ara, lo, nw)
	int *fn, *rc, *ara, *lo, *nw;
{
	*rc = 0==fread(ara+*lo-1, sizeof(*ara), *nw, IOtab[*fn].fptr);
}

getcsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	UNPACK *ara;
{
	register n;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	*rc = IOfget(fn, IOpbuf, n+1, IOtab[*fn].fptr)==EOF;
	if(*rc==1)
		return;
	if(strlen(IOpbuf) < n)
		IOpadbl(IOpbuf, n);
	IOunpack(IOpbuf, ara+*lo-1, n);
}


getwsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	PACK *ara;
{
	register n;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	*rc = IOfget(fn, IOpbuf, n+1, IOtab[*fn].fptr)==EOF;
	if(*rc==1)
		return;
	if(strlen(IOpbuf) < n)
		IOpadbl(IOpbuf, n);
	IOscramble(IOpbuf, ara+*lo-1, n);
}

getbsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	UNPACK *ara;
{
	register n;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	*rc = IOfget(fn, ara+*lo-1, n+1, IOtab[*fn].fptr)==EOF;
	if(*rc==1)
		return;
	*nc = strlen(IOpbuf);
	IOunpack(IOpbuf, ara+*lo-1, *nc);
}

getvsio(fn, rc, ara, lo, nc)
	int *fn, *rc, *lo, *nc;
	PACK *ara;
{
	register n;
	n = *nc==0? IOtab[*fn].linesize : *nc;
	*rc = IOfget(fn, ara+*lo-1, n+1, IOtab[*fn].fptr)==EOF;
	if(*rc==1)
		return;
	*nc = strlen(ara+*lo-1);
}

/* IOfget - fgets without trailing newline and with integer return */
IOfget(fn, s, n, iop)
	int *fn;
	char *s;
	register FILE *iop;
{
	register c;
	register char *cs;
	register IO *p;
	char *ttyname();
	p = &IOtab[*fn];
	if (p->access == GET  && p->prompting) { 
		/* here if possible prompt */
		if (isatty(fileno(p->fptr))) {
			fprintf(stderr, "%s", p->prompt);
		}
	}
	cs = s;
	while (--n>=0 && (c = getc(iop))>=0) {
		*cs = c;
		if (c=='\n')
			break;
		cs++;
	}
	while(c!='\n' && c>=0)
		c=getc(iop);
	if (c<0 && cs==s)
		return(EOF);
	*cs = '\0';
	return(0);
}

IOpack(a, b, n)
	register UNPACK *a;
	register char *b;
	register n;
{
	while(n--)
		*b++ = *a++;
}

IOscramble(c, i, n)
	register char *c;
	register SDS *i;
	register n;
{
	char w[4];
	do{
		w[3] = *c++;
		w[2] = *c++;
		w[1] = *c++;
		w[0] = *c++;
		*i++ = ((int *)w)[0];
	}while((n-=4)>0);
}
IOunpack(a, b, n)
	register char *a;
	register UNPACK *b;
	register n;
{
	while(n--)
		*b++ = *a++&0377;
}

IOstripbl(s, n)
	register char *s;
	register n;
{
	register char *p;
	for(p=s+n-1; *p==' ' || *p=='\t'; )
		if(--p==s-1)
			break;
	*++p = '\n';
	*++p = '\0';
}

IOpadbl(s, n)
	register char *s;
	register n;
{
	register char *p;
	p=s+strlen(s);
	while(p < s+n)
		*p++ = ' ';
}

lntime(ara)
	int ara[8];
{
	long	time(), t;
#include <sys/time.h>
	struct tm *tm, *localtime();
	t = time(0);
	tm = localtime(&t);
	ara[0] = tm->tm_year+1900;
	ara[1] = tm->tm_mon+1;
	ara[2] = tm->tm_mday;
	ara[3] = tm->tm_hour;
	ara[4] = tm->tm_min;
	ara[5] = tm->tm_sec;
	ara[6] = tm->tm_yday;
	ara[7] = tm->tm_wday+1;
}

letime(t)
	int *t;
{
#include <sys/times.h>
#define	HZ	60
	struct tms tms;
	times(&tms);
	*t=((tms.tms_utime+tms.tms_stime)*1000)/HZ;
}

ltlxtr()
{
}

sysini()
{
}

sysfin(termcode, compcode)
	int *termcode, *compcode;
/* assign error codings compatible with little.
 *	termcode=0,compcode=0 -> exit 0
 *	termcode=0,compcode=4 -> exit 0 (but is 'warning' state)
 *	termcode=0,compcode=8 -> exit 88 ('error' state)
 *	termcode=0,compcode=0 -> exit 90
 *	termcode=1,compcode 1..9 -> exit 91..99
 *	termcode 1001..1019 -> exit 101...119 (little library error)
 *	termcode=1100..1199 -> NEVER OCCURS (math library in c)
 *	termcode 1201..1231 -> exit 121..151 (little multi-word error)
 *	termcode=1301..1322 -> exit 161..182 (little i/o error)
 */
{
	if(*termcode==0 && *compcode==0)
		exit(0);
	else if (*termcode==0 && *compcode==8) /* if errors */
		exit(88);
	else if (*termcode==0)
		exit(0);
	else if (*termcode==1 && *compcode==0)
		exit(91);
	else if (*termcode==1 && *compcode<=9)
		exit(90+*compcode);
	else if (*termcode==1 && *compcode>=1001 && *compcode<=1019)
		exit(100+(*compcode-1000));
	else if (*termcode==1 && *compcode >= 1201 && *compcode<=1231)
		exit(120+(*compcode-1200));
	else if (*termcode==1 && *compcode >=1301 && *compcode <= 1322)
		exit(160+(*compcode-1300));
	else /* all others taken as 1,1 */
		exit(91);

}

char IOav[256];
char *IOavp=IOav;

main(argc, argv) 
	char *argv[];
{
/* bypass tty code (21-apr-82)
	tty = open("/dev/tty", 1);
	if(tty < 0){
		perror("tty open");
		exit(1);
	}
*/
	while(--argc){
		strcpy(IOavp, *++argv);
		IOavp+=strlen(IOavp);
		*IOavp++=',';
	}
	*--IOavp='\0';
	_ltlmai();
}

readsos(ara, ptr)
	UNPACK *ara;
	int *ptr;
{
	*ptr = strlen(IOav);
	IOunpack(IOav, ara, *ptr);
}

remarkl(ara)
	char *ara;
{
/* bypass tty code (21-apr-82)
	IOsdstoc(ara, IOpbuf);
	write(tty, IOpbuf, strlen(IOpbuf));
	write(tty, "\n", 1);
*/
}
eretsio(){}
namesio(fn, rc, s, sl)
	int *fn;
	int *rc;
	SDS *s;
	int *sl;
{ /*rc=0 (normal) 1 (name truncated to *sl chars) 2 (error) */
	*rc=2;
}
ecodsio(){}
promsio(fn, rc, str)
	int *fn, *rc;
	SDS *str;
/* if file fn opened for GET access and is terminal, set prompt 
 * string to str. 
*/
{
	register IO *p;

	*rc = 0;
	p = &IOtab[*fn];
	p->prompting = 0; /* turn off prompting */
	if (p->access != GET) return;
	IOsdstoc(str, p->prompt);
	if (p->prompt[0]=='\0') return;
	if (p->prompt[0]=='0' && p->prompt[1]=='\0') return;
	p->prompting = 1;
}
/*JUNK
*ltlini(){}
*fclose(){}
*ltlfin(){}
*IOsdstoc(){}
*/
