#
# These routines are used to load and store absolute addressed within
# the image.
#
# Author:  D. Shields  (NYU-CIMS)  15-Sep-81
# Revise:  D. Shields  (NYU-CIMS)  05-Feb-82
#		Convert from VMS macro to Unix as, as part of initial
#		port.
#
#
# Entry points:
#
#	PICALL		call a subroutine
#	PIMADR		get memory address
#	PICRSD		create hardware string descriptor
#	PIGETB		get byte
#	PIGETW		get word
#	PIPUTB		put byte
#	PIPUTW		put word
#
	.text
	.align	1

	.globl	picall
picall:	.word	0x0	# m<>
#
# This routine calls the procedure specified by its first argument
# with the parameter list specified by its second argument.
#
#	call PICALL(EXTADR, ARGADR)
# where:
#	EXTADR	entry address of external procedure
#	ARGADR	address of parameter list
#
	movl	*4(ap), r1	# move EXTADR into r1
	movl	*8(ap), r0	# move ARGADR into r0
	callg	(r0), (r1)	# issue call
	ret			# done


	.globl	pimadr
pimadr:	.word	0x0	# m<>
#
# This routine returns the (byte) address of its argument.
#
	movl	4(ap), r0	# get address.
	ret			# done.


	.globl	picrsd
picrsd:	.word	0x000c	# m<r2,r3>
#
# This routine determines starting address of string referenced by
# the second argument string specifier, and stores the address at the
# heap location whose index is given by the first argument.
#
#	call PIMFSD(T, SPEC, HEAP)#
# where:
#	T	heap index of word to receive starting address
#		of string
#	SPEC	a SETL specifier for a long SETL string.
#	HEAP	base address of heap (i.e. address of HEAP(1))
#
# SETL string specifier assume the following declarations:
#	VALUE_   =  .f. 03, 22,
#	IC_OFS_  =  .f. 01, 16,		OFF_IC_LEN  =  01
#	IC_LEN_  =  .f. 17, 16,		OFF_IC_PTR  =  01
#	IC_PTR_  =  .f. 03, 22,		OFF_IC_PTR  =  02
 
# just set address of string (DS 6-feb-82)
	movl	12(ap), r1		# base address for HEAP(1)
	extzv	$2, $22, *8(ap), r0	# index of IC_BLK
	moval	(r1)[r0], r2		# address of 2nd word of IC_BLK
	movl	*4(ap), r3		# load heap index
	decl	r3			# account for one-s origin
	extzv	$2, $22, 4(r2), r0	# get IC_PTR
	moval	4(r1)[r0], r0		# point to first word
	movzwl	(r2), r2		# get IC_OFS
	addl3	r0, r2, (r1)[r3]	# store address
	ret

# original vms code transcribed to unix as follows below:
	movl	12(ap), r1		# base address for HEAP(1)
	extzv	$2, $22, *8(ap), r0	# index of IC_BLK
	moval	(r1)[r0], r2		# address of 2nd word of IC_BLK
	movl	*4(ap), r3		# address of VAX string desc

	#movw	2(r2), dsc$w_length(r3)	# move length
	movw	2(r2),(r3)		# move length

	extzv	$2, $22, 4(r2), r0	# get IC_PTR
	moval	4(r1)[r0], r0		# point to first word
	movzwl	(r2), r2		# get IC_OFS
	#addl3	r0, r2, dsc$a_pointer(r3)	# set address
	addl3	r0, r2, 4(r3)		# (above in as).
	ret				# done


	.globl	pigetb
pigetb:	.word	0x0		# m<>
#
# This routine returns the contents of the byte referenced by its
# argument.
#
	movl	*4(ap), r0	# get byte index.
	movzbl	(r0), r0 	# load desired byte.
	ret			# done.


	.globl	pigetw
pigetw:	.word	0x0		# m<>
#
# This routine returns the contents of the word referenced by its
# argument.
#
	movl	*4(ap), r0	# get byte index.
	movl	(r0), r0	# load desired word.
	ret			# done.


	.globl	piputb
piputb:	.word	0x0		# m<>
#
# This routine stores its second argument byte in the address 
# referenced by its first argument.
#
	movl	*4(ap), r0	# get address.
	movb	*8(ap), (r0)	# store the word.
	ret			# done.


	.globl	piputw
piputw:	.word	0x0		# m<>
#
# This routine stores its second argument word in the address 
# referenced by its first argument.
#
	movl	*4(ap), r0	# get address.
	movl	*8(ap), (r0)	# store the word.
	ret			# done.

	.globl	_STLINT
_STLINT:
	.word	0x0		# m<>
	pushl	8(ap)
	pushl	4(ap)
	calls	$2,stlint
	ret
